# -----------------------------------------------------------------------------
# This script file provides fundamental computational functionality for the
# the development of a quadratic classifier. The learning process will be 
# implemented in the context of a binary classification problem where patterns
# from each class will follow a multi-dimensional gaussian distribution.
# -----------------------------------------------------------------------------
#                                                        T 
# Assume that each training pattern x = [x_1,x_2,...,x_d]  is d-dimensional.
# -----------------------------------------------------------------------------
#                For each x in C1 ==> x ~ N_d(MU1,SIGMA1)
#                For each x in C2 ==> x ~ N_d(MU2,SIGMA2)
# -----------------------------------------------------------------------------
# The decision surface between the two classes C1 and C2 to be learned is 
# assumed to be of quadratic form such that:
#
#                            T            T
#           f(x;W2,W1,W0) = x * W2 * x + x * W1 + W0 = 0
# -----------------------------------------------------------------------------

# Import required Python libraries.
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from torch.utils.data import Dataset, DataLoader
import matplotlib.pyplot as plt
# -----------------------------------------------------------------------------
# STEP 0: SET RANDOM SEEDS FOR REPRODUCIBILITY
# -----------------------------------------------------------------------------

torch.manual_seed(0)
np.random.seed(0)

# Define a function to get the correct training environemnt for the model.
def get_execution_device():
    # Set the existence status of a mps GPU.
    if hasattr(torch.backends,"mps"):
        is_mps = torch.backends.mps.is_available()
    else:
        is_mps = False
    # Set the existence status of a cuda GPU.
    is_cuda = torch.cuda.is_available()
    # Check the existence status of a mps GPU to be used during training.
    if is_mps:
        device = torch.device("mps")
        print("MPS GPU is available!")
        print(70*"=")
    # Check the existence of a cuda GPU to be used during training.
    elif is_cuda:
        device = torch.device("cuda")
        print("CUDA GPU is available!")
        print(70*"=")
    # Otherwise, a CPU device will be used instead.
    else:
        device = torch.device("cpu")
        print("GPU is not available, CPU will be used instead!")
        print(70*"=")
    return device

# -----------------------------------------------------------------------------
# STEP 1: CUSTOM DATASET CLASS DEFINITION FOR SYNTHETIC DATA GENERATION
# -----------------------------------------------------------------------------

class MultivariateGaussianDataset(Dataset):
    
    # Class Constructor
    def __init__(self,n_samples,mean1,mean2,cov,train_ratio,split,device):
        # Set the execution device.
        self.device = device
        # Set the splitting mode. If it is not equal to train it will be 
        # considered as test.
        self.split = split
        # Initialize the two probability distribution objects.
        dist1 = torch.distributions.MultivariateNormal(mean1,cov)
        dist2 = torch.distributions.MultivariateNormal(mean2,cov)
        # Generate the actual data patterns for each class.
        data1 = dist1.sample((n_samples,)).to(self.device)
        data2 = dist2.sample((n_samples,)).to(self.device)
        # Generate the corresponding target labels.
        labels1 = torch.zeros((n_samples,1),device=self.device) # 0 for Class 1
        labels2 = torch.ones((n_samples,1),device=self.device)  # +1 for Class 2
        # Combine data patterns and associated target labels to corresponding 
        # tensors.
        self.data = torch.cat((data1,data2),dim=0)
        self.labels = torch.cat((labels1,labels2),dim=0)
        # Initialize a permutation tensor in order to perform an initial 
        # shuffling of the dataset. The number of elements in the permutation
        # tensor is equal to the number of rows of the data and labels tensors.
        perm = torch.randperm(self.data.size(0),device=self.device)
        # Apply the permutation tensor to both the data and labels tensors.
        self.data, self.labels = self.data[perm], self.labels[perm]
        # Set the split index that will eventually partition the data and labels
        # tensors into training and testing subsets.
        split_index = int(n_samples * 2 * train_ratio)
        if split == "train":
            self.data = self.data[:split_index]
            self.labels = self.labels[:split_index]
        else:
            self.data = self.data[split_index:]
            self.labels = self.labels[split_index:]
    
    # Length Method
    def __len__(self):
        return len(self.data)
    
    # Slicing Enabler Method
    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]

# -----------------------------------------------------------------------------
# STEP 2: INSTANTIATE TRAIN AND TESTING DATASET OBJECTS AND DATALOADERS
# -----------------------------------------------------------------------------

# Set the device on which training will be performed.
device = get_execution_device()

# Define fundamental simulation parameters.
input_dim = 2
n_samples = 500
displacement = 5
cov_scale = 1
train_ratio = 0.8
batch_size = 50

# Set the locations of the two mean vectors.
mean1 = torch.randn(input_dim)
mean2 = torch.randn(input_dim)+displacement
# Set the covariance matrix to be used for both probability density 
# functions.
cov = torch.eye(input_dim) * cov_scale

# Instantiate the train dataset object.
train_dataset = MultivariateGaussianDataset(n_samples,mean1,mean2,cov,
                                            train_ratio,split="train",
                                            device=device)
# Instantiate the test dataset object.
test_dataset = MultivariateGaussianDataset(n_samples,mean1,mean2,cov,
                                            train_ratio,split="test",
                                            device=device)

# Instantiate the train and test loader objects.
# The DataLoader class can be configured to shuffle the data by setting the 
# shuffle argument to True when creating a DataLoader object. When shuffling is 
# enabled, the DataLoader class randomly shuffles the indices of the data points 
# and returns the data in a random order for each epoch. Apparently, shuffling 
# is not enabled during testing since this process in executed only once.
train_loader = DataLoader(train_dataset,batch_size=batch_size,shuffle=True)
test_loader = DataLoader(test_dataset,batch_size=batch_size,shuffle=False) 

# -----------------------------------------------------------------------------
# STEP 3: QUADRATIC CLASSIFICATION MODEL
# -----------------------------------------------------------------------------

class QuadraticClassificationModel(nn.Module):
    
    # Class Constructor
    def __init__(self,input_dim,device):
        # Call the super class constructor.
        super(QuadraticClassificationModel,self).__init__()
        # Set the execution device.
        self.device = device
        # Intialize the weight parameter matrix W2 associated with the 
        # quadratic terms of the discrimination function.
        #                    T
        # Quadratic Term = x   * W2 * x
        self.W2 = nn.Parameter(torch.randn(input_dim,input_dim).to(self.device))
        # Initialize the weight parameter vector W1 associated with the 
        # linear terms of the discrimination function.
        #                 T
        # Linear Term = x * W1
        self.W1 = nn.Parameter(torch.randn(input_dim).to(self.device))
        # Initialize the weight parameter scalar W0 associated with the
        # bias term of the discrimination function.
        # Bias Term: W0
        self.W0 = nn.Parameter(torch.tensor(0.0).to(self.device))
        
    # Forward Pass
    def forward(self,x):
        # Input tensor x is assummed to have a (batch_size,input_dim) shape.
        # Ensure that the input tensor is moved to the correct device.
        x = x.to(self.device)
        #                                     T 
        # Quadratic term computation: Q(x) = x * W2 * x
        # involves a tensor operation that must account for the batch dimension. 
        # The operation computes the quadratic form for each sample in the batch, 
        # where each sample is represented as a vector of size [input_dim]
        quadratic_term = torch.einsum('bi,ij,bj->b', x, self.W2, x)
        # Using the Einstein summation in the previous way ensures that the
        # quadratic form is correctly applied to each sample in the batch, 
        # producing a single scalar for each sample that represents the result
        # of the quadratic term
        #                                  T
        # Linear term computation: L(x) = x * W1
        # is more straightforward because it involves a matrix-vector product 
        # for each sample in the batch with the weight vector W1.
        linear_term = torch.matmul(x, self.W1)
        # Bias term computation: B(x) = W0
        # is even simpler since it involves adding the bias weigth term W0 to 
        # each sample's result in the batch.
        bias_term = self.W0.unsqueeze(0)
        return torch.sigmoid(quadratic_term+linear_term+bias_term)
    
# -----------------------------------------------------------------------------
# STEP 4: MODEL TRAINING AND EVALUATION METHODS
# -----------------------------------------------------------------------------

# Training function definition.

def train_model(model,train_loader,test_loader,criterion,optimizer,device):
    
    # The training process will continue to be executed until the training 
    # loss of the next epoch is greater than the training loss of the previous
    # epoch.
    
    # Initialize the previous epoch loss with infinity.
    previous_epoch_loss = float('inf')  
    # Initialize the epoch counter.
    epoch = 0
    # Loop uniil the training loss decreases.
    while True:
        # Set the training execution environment.
        model.train()
        # Initialize variables to be used to measure the average training 
        # accuracy and loss for each epoch.
        running_loss = 0.0
        correct_predictions = 0
        total_predictions = 0
        # Loop through the various training batches of inputs and their 
        # corresponding labels.
        for inputs,labels in train_loader:
            # Set the gradient vector associated with the internal weight 
            # parameters of the model to zero.
            optimizer.zero_grad()
            # Acquire estimated outputs of the model on the current batch of 
            # input patterns.
            outputs = model(inputs)
            # Evaluate the loss function on the current pair of true and 
            # estimated labels.
            loss = criterion(outputs.view(-1,1),labels)
            # Update the running loss estimate.
            running_loss += loss.item()
            # Run the backward propagation of information within the model.
            loss.backward()
            # Upadate the internal weight parameters of the network.
            optimizer.step()
            # Get the number of predictions mades during processing the current
            # batch.
            total_predictions += labels.size(0)
            # Calculate accuracy at the current batch level.
            predicted = outputs.round()
            correct_predictions += (predicted.view(-1,1)==labels).sum().item()
        
        # Compute the average loss and accuracy over the epoch.
        epoch_loss = running_loss / len(train_loader)
        epoch_accuracy = correct_predictions / total_predictions
        # Evaluate the testing loss of the model.
        test_loss,test_accuracy = evaluate_accuracy(model,test_loader,criterion)
        # Report accuracy measurements.
        report_string = f"Epoch {epoch+1}, Training Loss: {epoch_loss:.8f},"
        report_string += f"Training Accuracy: {epoch_accuracy:.2f},"
        report_string +=  f"Testing Accuracy: {test_accuracy:.2f}"
        print(report_string)
        
        # Check if the current epoch loss is greater than the previous epoch loss
        if epoch_loss > previous_epoch_loss:
            print("Stopping training as the loss increased from the previous epoch.")
            break  # Stop the training process
            
        # Update the previous_epoch_loss with the current epoch_loss
        previous_epoch_loss = epoch_loss
        epoch += 1

# Testing function definition.
def evaluate_accuracy(model,data_loader,criterion):
    
    # Set the testing execution environment.
    model.eval()
    # Initalize correct and total predictions countersn and running loss.
    correct_predictions = 0
    total_predictions = 0
    running_loss = 0.0
    
    # Loop through the various testing batches but without performing parameters
    # update.
    with torch.no_grad():
        for inputs,labels in data_loader:
            # Aquire estimated outputs of the model on the current batch of 
            # input patterns.
            outputs = model(inputs)
            # Evaluate the loss function on the current pair of true and 
            # estimated labels.
            loss = criterion(outputs.view(-1,1),labels)
            # Update the running loss estimate.
            running_loss += loss.item()
            # Get the number of predictions mades during processing the current
            # batch.
            total_predictions += labels.size(0)
            # Calculate accuracy at the current batch level.
            predicted = outputs.round()
            correct_predictions += (predicted.view(-1,1)==labels).sum().item()
    # Compute the average loss and accuracy over the epoch.
    loss = running_loss / len(train_loader)
    accuracy = correct_predictions / total_predictions
    return loss,accuracy

# -----------------------------------------------------------------------------
# STEP 5: MODEL INITIALIZATION, TRAINING AND EVALUATION
# -----------------------------------------------------------------------------

# Initialize model.
model = QuadraticClassificationModel(input_dim,device)
# Move the initialized model to the appropriate device.
model = model.to(device)
# Set the Mean Squared Error as the criterion to be optimized.
criterion = nn.BCELoss()
# Set the optimizer to be used throuout training.
# Set the learning rate parameter to 0.01
optimizer = optim.Adam(model.parameters(),lr=0.01)

# Train and evaluate the model.
train_model(model,train_loader,test_loader,criterion,optimizer,device)

# -----------------------------------------------------------------------------
# STEP 6: MODEL VISUALIZATION
# -----------------------------------------------------------------------------

# Auxiliary function for grouping training and testing patterns according 
# to class. In the context of this example, patterns from two distinct classes
# are considered.
def group_data(data_loader):
    
    # Initialize list containers to store patterns from the different classes.
    class1_patterns = []
    class2_patterns = []
    
    # Iterate through the data loader.
    for inputs,labels in data_loader:
        # Isolate and store data patterns originating from class 1 (0).
        class1_patterns.append(inputs[labels.squeeze()==0])
        # Isolate and store data patterns originating from class 2 (1).
        class2_patterns.append(inputs[labels.squeeze()==1])
    
    # Concatenate lists into tensors and ensure  that all tensors are 
    # transfered on the cpu before plotting.
    class1_patterns = torch.cat(class1_patterns,dim=0).cpu()
    class2_patterns = torch.cat(class2_patterns,dim=0).cpu()
    
    return class1_patterns, class2_patterns
    

# Plotting function for two-dimensional data patterns.
def plot_data(model,train_loader,test_loader,device):
    
    # Get the dimensionality of the input data.
    input_dim = train_loader.dataset.data.size(1)
    
    # Plotting will only be performed for two-dimensional data.
    if input_dim != 2:
        print("Visualization is only supported for 2D data.")
        return
    
    # Organize training and testing data by class.
    train_class1_patterns, train_class2_patterns = group_data(train_loader)
    test_class1_patterns, test_class2_patterns = group_data(test_loader)
    
    # Initialize plotting figure and system of coordinates.
    fig, ax = plt.subplots()
    
    # Adjust the point size.
    point_size = 20
    
    # Plot training data by class
    ax.scatter(train_class1_patterns[:,0],train_class1_patterns[:,1],
               c='blue',label='Class A (Train)',alpha=0.6,s=point_size)
    ax.scatter(train_class2_patterns[:,0],train_class2_patterns[:,1], 
               c='red',label='Class B (Train)',alpha=0.6,s=point_size)
    
    # Plot testing data by class.
    ax.scatter(test_class1_patterns[:,0],test_class1_patterns[:,1],c='cyan',
               label='Class A (Test)',alpha=0.9,s=point_size)
    ax.scatter(test_class2_patterns[:,0],test_class2_patterns[:,1],c='orange',
               label='Class B (Test)', alpha=0.9, s=point_size)
        
    # Concatenate all data points to calculate the limits.
    all_data_points = torch.cat((train_class1_patterns,train_class2_patterns,
                                 test_class1_patterns,test_class2_patterns),dim=0)
    x_min, x_max = all_data_points[:, 0].min(), all_data_points[:, 0].max()
    y_min, y_max = all_data_points[:, 1].min(), all_data_points[:, 1].max()
    
    # Set the the global xy minimum and maximum in order to generate a Cartesian
    # system of coordinates where the two axes are equal.
    xy_min = min(x_min,y_min)
    xy_max = max(x_max,y_max)
    
    # Set the margin to add around the minimum and maximum values for clarity.
    margin = 0.1 
    
    # Generate a grid of points over the data range.
    # Define the linear space to be used for both the x and y axes based on 
    # the corresponding discretization value Nxy which provides the number of
    # points to be considered along each axes.
    Nxy = 100
    xy = np.linspace(xy_min-margin,xy_max+margin,Nxy)
    xx, yy = np.meshgrid(xy,xy)
    # Define the grid tensor by concatenating the one-dimensional (ravelled) 
    # contents of the np arrays xx and yy. Make sure that the tensor is moved
    # to the prespecified execution device.
    grid = torch.Tensor(np.c_[xx.ravel(),yy.ravel()]).to(device)
    
    # Predict over the grid without updating the internal weight parameters 
    # of the model.
    with torch.no_grad():
        predictions = model(grid).reshape(xx.shape)
        predictions = predictions.cpu().numpy()
    
    # Plot the decision boundary
    ax.contour(xx, yy, predictions, levels=[0.5], cmap="Greys", vmin=-0.5, vmax=0.5)
    
    # Set axes limits with the prespecified margin
    ax.set_xlim(xy_min - margin, xy_max + margin)
    ax.set_ylim(xy_min - margin, xy_max + margin)
    
    # Add figure title and x and y axes labels.
    plt.title('Training and Testing Data')
    plt.xlabel('Feature 1')
    plt.ylabel('Feature 2')
    
    # Add a grid to the plotting window.
    plt.grid(True)

    # Handle duplicate labels in the legend
    handles, labels = ax.get_legend_handles_labels() 
    unique_labels = dict(zip(labels, handles))
    ax.legend(unique_labels.values(), unique_labels.keys(),
              prop={"size":8},loc="lower right")
    plt.show()
    
# -----------------------------------------------------------------------------
# STEP 7: PERFORM MODEL VISUALIZATION 
# -----------------------------------------------------------------------------

# Call the plotting function.
plot_data(model,train_loader,test_loader,device=device)